# Generated by Django 4.1.3 on 2023-01-10 00:54

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ComponentParams",
            fields=[
                (
                    "uuid",
                    models.UUIDField(
                        editable=False, primary_key=True, serialize=False, unique=True
                    ),
                ),
                ("data", models.BinaryField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]
